import hashlib
import os
import re
from StringIO import StringIO
from zipfile import BadZipfile

from regex import run_regex_tests
from validator.constants import MAX_JS_THRESHOLD
from validator import decorator
from validator import submain as testendpoint_validator
from validator import unicodehelper
import validator.testcases.markup.markuptester as testendpoint_markup
import validator.testcases.markup.csstester as testendpoint_css
import validator.testcases.scripting as testendpoint_js
import validator.testcases.langpack as testendpoint_langpack
from validator.constants import (BUGZILLA_BUG, PACKAGE_LANGPACK,
                                 PACKAGE_SUBPACKAGE, PACKAGE_THEME)
from validator.xpi import XPIManager


FLAGGED_FILES = set(['.DS_Store', 'Thumbs.db'])
FLAGGED_EXTENSIONS = set(['.orig', '.old', '~'])
OSX_REGEX = re.compile('__MACOSX')

hash_library_allowed = {}
for hash_list in 'hashes-allowed.txt', 'static_hashes.txt':
    with open(os.path.join(os.path.dirname(__file__), hash_list)) as f:
        hash_library_allowed.update(s.strip().split(None, 1) for s in f)

hash_library_warning = {}
with open(os.path.join(os.path.dirname(__file__), "hashes-warning.txt")) as f:
        hash_library_warning.update(s.strip().split(None, 1) for s in f)


@decorator.register_test(tier=1)
def test_xpcnativewrappers(err, xpi_package=None):
    """Tests the chrome.manifest file to ensure that it doesn't contain
    xpcnativewrappers objects."""

    # Don't even both with the test(s) if there's no chrome.manifest.
    chrome = err.get_resource('chrome.manifest')
    if not chrome:
        return None

    for triple in chrome.triples:
        # Test to make sure that the triple's subject is valid
        if triple['subject'] == 'xpcnativewrappers':
            err.warning(('testcases_content',
                         'test_xpcnativewrappers',
                         'found_in_chrome_manifest'),
                        'xpcnativewrappers not allowed in chrome.manifest',
                        'chrome.manifest files are not allowed to contain '
                        'xpcnativewrappers directives.',
                        filename=triple['filename'],
                        line=triple['line'],
                        context=triple['context'])


@decorator.register_test(tier=2)
def test_packed_packages(err, xpi_package=None):
    'Tests XPI and JAR files for naughty content.'

    processed_files = 0
    pretested_files = err.get_resource('pretested_files') or []

    scripts = set()
    chrome = err.get_resource('chrome.manifest_nopush')
    overlays = chrome.get_applicable_overlays(err) if chrome else set()

    marked_scripts = err.get_resource('marked_scripts')
    if not marked_scripts:
        marked_scripts = set()

    identified_files = err.metadata.setdefault('identified_files', {})

    # Iterate each item in the package.
    for name in xpi_package:

        # Warn for things like __MACOSX directories and .old files.
        if '__MACOSX' in name or name.split('/')[-1].startswith('.'):
            err.warning(
                err_id=('testcases_content', 'test_packed_packages',
                        'hidden_files'),
                warning='Hidden files and folders flagged',
                description='Hidden files and folders complicate the review '
                            'process and can contain sensitive information '
                            'about the system that generated the XPI. Please '
                            'modify the packaging process so that these files '
                            "aren't included.",
                filename=name)
            continue
        elif (any(name.endswith(ext) for ext in FLAGGED_EXTENSIONS) or
              name in FLAGGED_FILES):
            err.warning(
                err_id=('testcases_content', 'test_packaged_packages',
                        'flagged_files'),
                warning='Flagged filename found',
                description='Files were found that are either unnecessary '
                            'or have been included unintentionally. They '
                            'should be removed.',
                filename=name)
            continue

        # Skip the file if it's in the pre-tested files resource. This skips
        # things like Jetpack files.
        if name in pretested_files:
            continue

        # Read the file from the archive if possible.
        file_data = u''
        try:
            file_data = xpi_package.read(name)
        except KeyError:  # pragma: no cover
            pass

        if not err.for_appversions:
            hash = hashlib.sha256(file_data).hexdigest()
            identified_warning = hash_library_warning.get(hash)
            if identified_warning is not None:
                identified_files[name] = {'path': identified_warning}
                err.warning(
                    err_id=('testcases_content',
                            'test_packed_packages',
                            'dangerous_js_library'),
                    warning='Dangerous JS Library Detected',
                    description=('Your add-on uses a JavaScript library we do '
                                 'not recommend. Read more: '
                                 'https://bit.ly/1TRIyZY',
                                 'File %r might be a dangerous or deprecated '
                                 'JavaScript library' % name),
                    filename=name)
                continue

            identified_allowed = hash_library_allowed.get(hash)
            if identified_allowed is not None:
                identified_files[name] = {'path': identified_allowed}
                err.notice(
                    err_id=('testcases_content',
                            'test_packed_packages',
                            'blacklisted_js_library'),
                    notice='JS Library Detected',
                    description=('JavaScript libraries are discouraged for '
                                 'simple add-ons, but are generally '
                                 'accepted.',
                                 'File %r is a known JS library' % name),
                    filename=name)
                continue

        # Process the file.
        processed = False
        name_lower = name.lower()
        if name_lower.endswith(('.js', '.jsm')):
            # Add the scripts to a list to be processed later.
            scripts.add(name)
        elif name_lower.endswith(('.xul', '.xml', '.html', '.xhtml', '.xbl')):
            # Process markup files outside of _process_file so we can get
            # access to information about linked scripts and such.
            parser = testendpoint_markup.MarkupParser(err)
            parser.process(name, file_data,
                           xpi_package.info(name)['extension'])
            run_regex_tests(file_data, err, name)

            # Make sure the name is prefixed with a forward slash.
            prefixed_name = name if name.startswith('/') else '/%s' % name
            # Mark scripts as pollutable if this is an overlay file and there
            # are scripts to mark.
            if overlays and prefixed_name in overlays and parser.found_scripts:
                # Look up the chrome URL for the overlay
                reversed_chrome_url = chrome.reverse_lookup(err, name)
                for script in parser.found_scripts:
                    # Change the URL to an absolute URL.
                    script = _make_script_absolute(reversed_chrome_url, script)
                    if script:
                        # Mark the script as potentially pollutable.
                        marked_scripts.add(script)
                        err.save_resource('marked_scripts', marked_scripts)
                    else:
                        err.warning(
                            err_id=('testcases_content',
                                    'test_packed_packages',
                                    'invalid_chrome_url'),
                            warning='Invalid chrome URL',
                            description='The referenced chrome: URL '
                                        'could not be resolved to a '
                                        'script file.',
                            filename=name)

        else:
            # For all other files, simply throw it at _process_file.
            processed = _process_file(err, xpi_package, name, file_data,
                                      name_lower)
            # If the file is processed, it will return True. If the process
            # goes badly, it will return False. If the processing is skipped,
            # it returns None. We should respect that.
            if processed is None:
                continue

        # This is tested in test_langpack.py
        if err.detected_type == PACKAGE_LANGPACK and not processed:
            testendpoint_langpack.test_unsafe_html(err, name, file_data)

        # This aids in creating unit tests.
        processed_files += 1

    # If there aren't any scripts in the package, just skip the next few bits.
    if not scripts:
        return processed_files

    # Save the list of scripts, along with where to find them and the current
    # validation state.
    existing_scripts = err.get_resource('scripts')
    if not existing_scripts:
        existing_scripts = []
    existing_scripts.append({'scripts': scripts,
                             'package': xpi_package,
                             'state': err.package_stack[:]})
    err.save_resource('scripts', existing_scripts)

    return processed_files


@decorator.register_test(tier=3)
def test_packed_scripts(err, xpi_package):
    """
    Scripts must be tested separately from normal files to allow for markup
    files to mark scripts as being potentially polluting.
    """

    # This test doesn't apply to subpackages. We keep references to the
    # subpackage bundles so we can process everything at once in an unpushed
    # state.
    if err.is_nested_package:
        return

    scripts = err.get_resource('scripts')
    if not scripts:
        return

    total_scripts = sum(len(bundle['scripts']) for bundle in scripts)
    exhaustive = True
    if total_scripts > MAX_JS_THRESHOLD:
        err.warning(
            err_id=('testcases_content', 'packed_js', 'too_much_js'),
            warning='TOO MUCH JS FOR EXHAUSTIVE VALIDATION',
            description='There are too many JS files for the validator to '
                        'process sequentially. An editor must manually '
                        'review the JS in this add-on.')
        exhaustive = False

    # Get the chrome manifest in case there's information about pollution
    # exemptions.
    chrome = err.get_resource('chrome.manifest_nopush')
    marked_scripts = err.get_resource('marked_scripts')
    if not marked_scripts:
        marked_scripts = set()

    # Process all of the scripts that were found seperately from the rest of
    # the package contents.
    for script_bundle in scripts:
        package = script_bundle['package']

        # Set the error bundle's package state to what it was when we first
        # encountered the script file during the content tests.
        for archive in script_bundle['state']:
            err.push_state(archive)

        for script in script_bundle['scripts']:
            file_data = unicodehelper.decode(package.read(script))

            run_regex_tests(file_data, err, script, is_js=True)
            # If we're not running an exhaustive set of tests, skip the full JS
            # parse and traversal.
            if not exhaustive:
                continue

            if marked_scripts:
                reversed_script = chrome.reverse_lookup(script_bundle['state'],
                                                        script)
                # Run the standard script tests on the script, but mark the
                # script as pollutable if its chrome URL is marked as being so.
                testendpoint_js.test_js_file(
                    err, script, file_data,
                    pollutable=reversed_script in marked_scripts)
            else:
                # Run the standard script tests on the scripts.
                testendpoint_js.test_js_file(err, script, file_data)

        for i in range(len(script_bundle['state'])):
            err.pop_state()


def is_mozilla_signed(xpi_package):
    return ('META-INF/manifest.mf' in xpi_package and
            'META-INF/mozilla.rsa' in xpi_package and
            'META-INF/mozilla.sf' in xpi_package)


def is_signed(xpi_package):
    return 'META-INF/manifest.mf' in xpi_package


@decorator.register_test(tier=2)
def test_signed_xpi(err, xpi_package):
    """Checks if XPI is signed.

    We don't want to specifically test for mozilla.* or zigbert.* filenames
    here, because the filenames could be just anything. Testing the presence of
    the manifest.mf file should be a good indicator that the file is signed."""
    if err.is_nested_package and (err.get_resource('is_multipackage') or
                                  xpi_package.filename.endswith('.xpi')):
        if not is_mozilla_signed(xpi_package):
            err.error(
                err_id=('testcases_content', 'unsigned_sub_xpi'),
                error='Sub-package must be signed',
                description='Add-ons which contain sub-packages must have all '
                            'sub-packages signed before uploading.')
    elif is_signed(xpi_package):
        err.warning(
            err_id=('testcases_content', 'signed_xpi'),
            warning='Package already signed',
            description='Add-ons which are already signed will be re-signed '
                        'when published on AMO. This will replace any '
                        'existing signatures on the add-on.')


def _process_file(err, xpi_package, name, file_data, name_lower,
                  pollutable=False):
    """Process a single file's content tests."""

    # If that item is a container file, unzip it and scan it.
    if name_lower.endswith('.jar'):
        # This is either a subpackage or a nested theme.
        is_subpackage = not err.get_resource('is_multipackage')
        # Unpack the package and load it up.
        package = StringIO(file_data)
        try:
            sub_xpi = XPIManager(package, mode='r', name=name,
                                 subpackage=is_subpackage)
        except BadZipfile:
            err.error(('testcases_content',
                       'test_packed_packages',
                       'jar_subpackage_corrupt'),
                      'Subpackage corrupt.',
                      'The subpackage appears to be corrupt, and could not '
                      'be opened.',
                      name)
            return None

        # Let the error bunder know we're in a sub-package.
        err.push_state(name)
        err.detected_type = (PACKAGE_SUBPACKAGE if is_subpackage else
                             PACKAGE_THEME)
        err.set_tier(1)
        supported_versions = (err.supported_versions.copy() if
                              err.supported_versions else
                              err.supported_versions)

        if is_subpackage:
            testendpoint_validator.test_inner_package(err, sub_xpi)
        else:
            testendpoint_validator.test_package(err, package, name)

        err.pop_state()
        err.set_tier(2)

        err.supported_versions = supported_versions

    elif name_lower.endswith('.xpi'):
        # It's not a subpackage, it's a nested extension. These are
        # found in multi-extension packages.

        # Unpack!
        package = StringIO(file_data)

        err.push_state(name_lower)
        err.set_tier(1)

        # There are no expected types for packages within a multi-
        # item package.
        testendpoint_validator.test_package(err, package, name)

        err.pop_state()
        err.set_tier(2)  # Reset to the current tier

    elif name_lower.endswith(('.css', '.js', '.jsm')):

        if not file_data:
            return None

        # Convert the file data to unicode
        file_data = unicodehelper.decode(file_data)
        is_js = name_lower.endswith(('.js', '.jsm'))

        if name_lower.endswith('.css'):
            testendpoint_css.test_css_file(err, name, file_data)

        elif is_js:
            testendpoint_js.test_js_file(err, name, file_data,
                                         pollutable=pollutable)

        run_regex_tests(file_data, err, name, is_js=is_js)

        return True

    else:
        if file_data:
            file_data = unicodehelper.decode(file_data)
            run_regex_tests(file_data, err, name, explicit=True)

    return False


def _make_script_absolute(xul_path, script):
    """Returns the absolute chrome URL for a script's URL."""

    # Ignore absolute URLs.
    if re.match(r'[a-z-]+:', script, re.IGNORECASE):
        return script

    if not xul_path:
        return

    if script.startswith('/'):
        xul_path_base = xul_path[9:]
        return 'chrome://%s%s' % (xul_path_base.split('/')[0], script)
    else:
        xul_path_split = xul_path.split('/')
        xul_path_split[-1] = script
        return '/'.join(xul_path_split)



import re

VERSION_PATTERN = re.compile('^[-+*.\w]{,32}$')


def validate_name(err, value, source):
    'Tests a manifest name value for trademarks.'

    ff_pattern = re.compile('(mozilla|firefox)', re.I)

    err.metadata['name'] = value

    if ff_pattern.search(value):
        err.warning(
            ('metadata_helpers', '_test_name', 'trademark'),
            'Add-on has potentially illegal name.',
            'Add-on names cannot contain the Mozilla or Firefox '
            'trademarks. These names should not be contained in add-on '
            'names if at all possible.',
            source)


def validate_id(err, value, source):
    'Tests a manifest UUID value'

    field_name = '<em:id>' if source == 'install.rdf' else 'id'

    id_pattern = re.compile(
        '('
        # UUID format.
        '\{[0-9a-f]{8}-([0-9a-f]{4}-){3}[0-9a-f]{12}\}'
        '|'
        # "email" format.
        '[a-z0-9-\.\+_]*\@[a-z0-9-\._]+'
        ')',
        re.I)

    err.metadata['id'] = value

    # Must be a valid UUID string.
    if not id_pattern.match(value):
        err.error(
            ('metadata_helpers', '_test_id', 'invalid'),
            'The value of {name} is invalid'.format(name=field_name),
            ['The values supplied for {name} in the {source} file is not a '
             'valid UUID string or email address.'.format(
                 name=field_name, source=source),
             'For help, please consult: '
             'https://developer.mozilla.org/en/install_manifests#id'],
            source)


def validate_version(err, value, source):
    'Tests a manifest version number'


    field_name = '<em:version>' if source == 'install.rdf' else 'version'

    err.metadata['version'] = value

    # May not be longer than 32 characters
    if len(value) > 32:
        err.error(
            ('metadata_helpers', '_test_version', 'too_long'),
            'The value of {name} is too long'.format(name=field_name),
            'Values supplied for {name} in the {source} file must be 32 '
            'characters or less.'.format(name=field_name, source=source),
            source)

    # Must be a valid version number.
    if not VERSION_PATTERN.match(value):
        err.error(
            ('metadata_helpers', '_test_version', 'invalid_format'),
            'The value of {name} is invalid'.format(name=field_name),
            'The values supplied for version in the {source} file is not a '
            'valid version string. It can only contain letters, numbers, and '
            'the symbols +*.-_.'.format(name=field_name, source=source),
            source)



import re
import types

from validator.python.HTMLParser import HTMLParser, HTMLParseError

import validator.testcases.scripting as scripting
import validator.unicodehelper as unicodehelper
from validator.testcases.markup import csstester
from validator.contextgenerator import ContextGenerator
from validator.constants import *

DEBUG = False

UNSAFE_TAGS = ('script', 'object', 'embed', 'base', )
UNSAFE_THEME_TAGS = ('implementation', 'browser', 'xul:browser', 'xul:script')
SELF_CLOSING_TAGS = ('area', 'base', 'basefont', 'br', 'col', 'frame', 'hr',
                     'img', 'input', 'li', 'link', 'meta', 'p', 'param', )
SAFE_IFRAME_TYPES = ('content', 'content-primary', 'content-targetable', )
TAG_NOT_OPENED = 'Tag (%s) being closed before it is opened.'

DOM_MUTATION_HANDLERS = (
        'ondomattrmodified', 'ondomattributenamechanged',
        'ondomcharacterdatamodified', 'ondomelementnamechanged',
        'ondomnodeinserted', 'ondomnodeinsertedintodocument', 'ondomnoderemoved',
        'ondomnoderemovedfromdocument', 'ondomsubtreemodified', )
UNSAFE_THEME_XBL = ('constructor', 'destructor', 'field', 'getter',
                    'implementation', 'setter', )

GENERIC_IDS = ('string-bundle', 'strings', )


REMOTE_URL_PATTERN = re.compile('(ht|f)tps?://')
_markedsectionclose = re.compile(r']\s*]\s*>')


class MarkupParser(HTMLParser):
    """Parse and analyze the versious components of markup files."""

    def __init__(self, err, strict=True, debug=False):
        HTMLParser.__init__(self)
        self.err = err
        self.is_jetpack = 'is_jetpack' in err.metadata  # Cache this value.
        self.line = 0
        self.strict = strict
        self.debug = debug

        self.context = None

        self.xml_state = []
        self.xml_line_stack = []
        self.xml_buffer = []
        self.xbl = False

        self.xml_state_scripts = []

        self.reported = set()
        self.found_scripts = set()  # A set of script URLs in the doc.

        # Added as a patch for various Python HTMLParser issues.
        self.cdata_tag = None

    def process(self, filename, data, extension='xul'):
        """Processes data by splitting it into individual lines, then
        incrementally feeding each line into the parser, increasing the
        value of the line number with each line."""

        self.line = 0
        self.filename = filename
        self.extension = extension.lower()

        self.reported = set()

        self.context = ContextGenerator(data)

        lines = data.split('\n')

        buffering = False
        for line in lines:
            self.line += 1

            search_line = line
            while True:
                # If a CDATA element is found, push it and its contents to the
                # buffer. Push everything previous to it to the parser.
                if '<![CDATA[' in search_line and not buffering:
                    # Find the CDATA element.
                    cdatapos = search_line.find('<![CDATA[')

                    # If the element isn't at the start of the line, pass
                    # everything before it to the parser.
                    if cdatapos:
                        self._feed_parser(search_line[:cdatapos])
                    # Collect the rest of the line to send it to the buffer.
                    search_line = search_line[cdatapos + 9:]
                    buffering = True
                    continue

                elif ']]>' in search_line and buffering:
                    # If we find the end element on the line being scanned,
                    # buffer everything up to the end of it, and let the rest
                    # of the line pass through for further processing.
                    end_cdatapos = search_line.find(']]>')
                    self._save_to_buffer(search_line[:end_cdatapos])
                    search_line = search_line[end_cdatapos + 3:]
                    buffering = False
                break

            if buffering:
                self._save_to_buffer(search_line + '\n')
            else:
                self._feed_parser(search_line)

    def _feed_parser(self, line):
        """Feed incoming data into the underlying HTMLParser."""

        line = unicodehelper.decode(line)

        try:
            try:
                self.feed(line + '\n')
            except UnicodeDecodeError:
                line = line.decode('ascii', 'ignore')
                self.feed(line + '\n')

        except HTMLParseError as err:
            if DEBUG:  # pragma: no cover
                print self.xml_state, err

            if 'markup' in self.reported:
                return

            if self.strict:
                self.err.warning(
                    err_id=('markup', '_feed', 'parse_error'),
                    warning='Markup parsing error',
                    description=('There was an error parsing a markup '
                                 'file.', str(err)),
                    filename=self.filename,
                    line=self.line,
                    context=self.context)
            self.reported.add('markup')

    def handle_startendtag(self, tag, attrs):
        self.handle_starttag(tag, attrs, True)
        self.handle_endtag(tag)

    def handle_starttag(self, tag, attrs, self_closing=False):

        # Normalize!
        tag = tag.lower()
        # XUL scripts are identical to normal scripts. Treat them the same.
        if tag == 'xul:script':
            tag = 'script'

        orig_tag = tag

        # Be extra sure it's not a self-closing tag.
        if not self_closing:
            self_closing = tag in SELF_CLOSING_TAGS

        if DEBUG:  # pragma: no cover
            print 'S: ', self.xml_state, tag, self_closing

        # A fictional tag for testing purposes.
        if tag == 'xbannedxtestx':
            self.err.error(
                err_id=('markup', 'starttag', 'banned_element'),
                error='Banned markup element',
                description='A banned markup element was found.',
                filename=self.filename,
                line=self.line,
                context=self.context)

        # Test for banned XBL in themes.
        if self.err.detected_type == PACKAGE_THEME:
            if tag.startswith('xbl:'):
                self.xbl = True
                tag = tag[4:]

            # Find XBL elements.
            if self.xbl:
                if tag in UNSAFE_THEME_XBL:
                    self.err.warning(
                        err_id=('markup', 'starttag',
                                'unsafe_theme_xbl_element'),
                        warning='Banned XBL element in theme.',
                        description=('Certain XBL elements are disallowed in '
                                     'full themes.',
                                     'Element: <xbl:%s>' % tag),
                        filename=self.filename,
                        line=self.line,
                        context=self.context)

                elif (tag == 'property' and
                      any(a[0] in (u'onset', u'onget') for a in attrs)):
                    self.err.warning(
                        err_id=('markup', 'starttag',
                                'theme_xbl_property'),
                        warning='Themes are not allowed to use XBL properties',
                        description='XBL properties cannot be used in themes.',
                        filename=self.filename,
                        line=self.line,
                        context=self.context)

        # Test for banned elements in language pack and theme markup.
        if self.err.detected_type in (PACKAGE_LANGPACK, PACKAGE_THEME):
            if (tag in UNSAFE_TAGS or
                (self.err.detected_type == PACKAGE_THEME and
                 tag in UNSAFE_THEME_TAGS)):
                self.err.warning(
                    err_id=('markup', 'starttag', 'unsafe_langpack_theme'),
                    warning='Unsafe tag for add-on type',
                    description=('A tag in your markup has been marked as '
                                 'being potentially unsafe. Consider '
                                 'alternate means of accomplishing what the '
                                 'code executed by this tag performs.',
                                 'Tag "%s" is disallowed.' % tag),
                    filename=self.filename,
                    line=self.line,
                    context=self.context)

            # Make sure all src/href attributes are local
            if any(not self._is_url_local(attr[1]) for attr in attrs if
                   attr[0] in ('src', 'href')):
                self.err.warning(
                    err_id=('markup', 'starttag', 'remote_src_href'),
                    warning='`src`/`href` attributes must be local.',
                    description='Full Themes and language packs may not '
                                'reference remote resources.',
                    filename=self.filename,
                    line=self.line,
                    context=self.context)

        if tag == 'prefwindow':
            # Flag <prefwindow> elements without IDs.

            if not any((key == 'id') for key, val in attrs):
                self.err.warning(
                        err_id=('markup', 'starttag', 'prefwindow_id'),
                        warning='`<prefwindow>` elements must have IDs.',
                        description='`<prefwindow>` elements without `id` '
                                    'attributes cause errors to be reported '
                                    'in the error console and prevent '
                                    'persistence of certain properties of the '
                                    'dialog.',
                        filename=self.filename,
                        line=self.line,
                        context=self.context)

        elif tag in ('iframe', 'browser') and self.extension == 'xul':
            # Bork if XUL iframe has no type attribute

            type_ = None
            src = None
            for attr in attrs:
                attr_name = attr[0].lower()
                if attr_name == 'type':
                    type_ = attr[1].lower()
                elif attr_name == 'src':
                    src = attr[1].lower()

            # We say it's true by default to catch elements that are
            # type="chrome" without an src="" attribute.
            remote_src = True
            if isinstance(src, types.StringTypes):
                remote_src = not self._is_url_local(src)

            if (type_ and
                not (type_ in SAFE_IFRAME_TYPES or
                     not remote_src)):
                self.err.warning(('markup', 'starttag', 'iframe_type_unsafe'),
                                 'iframe/browser missing or unsafe `type` '
                                 'attribute',
                                 'All iframe and browser elements must have '
                                 'either a valid and safe `type` attribute or a'
                                 '`src` attribute that points to a local file.',
                                 self.filename,
                                 line=self.line,
                                 context=self.context)
            elif ((not type_ or
                   type_ not in SAFE_IFRAME_TYPES) and
                  remote_src):
                self.err.warning(('markup', 'starttag', 'iframe_type_unsafe'),
                                 'Typeless iframes/browsers must be local.',
                                 'iframe and browser elements that lack a type '
                                 'attribute must always have src attributes '
                                 'that reference local resources.',
                                 self.filename,
                                 line=self.line,
                                 context=self.context)

        elif tag == 'script':
            # Per the Addon Validator Spec (v2), scripts
            # must not be remote.

            src = None
            for attr in attrs:
                if attr[0].lower() == 'src':
                    src = attr[1].lower()
                    break

            if src:
                if not self._is_url_local(src):
                    self.err.warning(
                        err_id=('markup', 'starttag',
                                'banned_remote_scripts'),
                        warning='Scripts must not be remote',
                        description='<script> tags must not be referenced to '
                                    'script files that are hosted remotely.',
                        signing_help='Please do not attempt to load remote '
                                     'scripts into any privileged contexts. '
                                     'If you cannot avoid using remote '
                                     'scripts, please consider loading a '
                                     'remote document into an iframe, and '
                                     'allow that document to load the '
                                     'remote scripts that you need.',
                        signing_severity='high',
                        filename=self.filename,
                        line=self.line,
                        context=self.context)
                else:
                    self.found_scripts.add(src)

        # Find CSS and JS attributes and handle their values like they
        # would otherwise be handled by the standard parser flow.
        for attr in attrs:
            attr_name, attr_value = attr[0].lower(), attr[1]

            # We don't care about valueless attributes.
            if attr_value is None:
                continue

            if (attr_name == 'xmlns:xbl' and
                attr_value == 'http://www.mozilla.org/xbl'):
                self.xbl = True

            # Test that an absolute URI isn't referenced in Jetpack 1.4.
            if (self.is_jetpack and
                attr_value.startswith('resource://') and
                '-data/' in attr_value):
                self.err.warning(
                        err_id=('markup', 'starttag',
                                'jetpack_abs_uri'),
                        warning='Absolute URI referenced in Jetpack 1.4',
                        description=('As of Jetpack 1.4, absolute URIs are no '
                                     'longer allowed within add-ons.',
                                     'See %s for more information.' %
                                         JETPACK_URI_URL),
                        filename=self.filename,
                        line=self.line,
                        context=self.context,
                        compatibility_type='error')

            if (self.err.detected_type == PACKAGE_THEME and
                attr_value.startswith(('data:', 'javascript:'))):

                self.err.warning(
                        err_id=('markup', 'starttag',
                                'theme_attr_prefix'),
                        warning='Attribute contains banned prefix',
                        description=("A mark element's attribute contains a "
                                     'prefix which is not allowed in full '
                                     'themes.',
                                     'Attribute: %s' % attr_name),
                        filename=self.filename,
                        line=self.line,
                        context=self.context)

            if attr_name == 'style':
                csstester.test_css_snippet(self.err,
                                           self.filename,
                                           attr_value,
                                           self.line)
            elif attr_name.startswith('on'):  # JS attribute
                # Warn about DOM mutation event handlers.
                if attr_name in DOM_MUTATION_HANDLERS:
                    self.err.warning(
                        err_id=('markup', 'starttag',
                                'dom_manipulation_handler'),
                        warning='DOM Mutation Events Prohibited',
                        description='DOM mutation events are deprecated and '
                                    'have severe performance implications. '
                                    'Please use mutation observers instead: '
                                    'https://developer.mozilla.org/docs/Web/'
                                    'API/MutationObserver',
                        filename=self.filename,
                        line=self.line,
                        context=self.context)

                scripting.test_js_snippet(
                    err=self.err, data=attr_value,
                    filename=self.filename, line=self.line,
                    context=self.context)


            # Test for generic IDs
            if attr_name == 'id' and attr_value in GENERIC_IDS:
                self.err.warning(
                    err_id=('markup',
                            'starttag', 'generic_ids'),
                    warning='Overlay contains generically-named IDs',
                    description='An overlay is using a generically-named ID '
                                'that could cause compatibility problems with '
                                'other add-ons. Add-ons must namespace all IDs '
                                'in the overlay, in the same way that '
                                'JavaScript objects must be namespaced.',
                    filename=self.filename,
                    line=self.line,
                    context=self.context)

        if tag == 'script':
            self.xml_state_scripts.append(any(
                (x[0] == 'type' and 'javascript' not in x[1]) for x in attrs))

        # When the dev forgets their <!-- --> on a script tag, bad
        # things happen.
        if 'script' in self.xml_state and tag != 'script':
            self._save_to_buffer('<' + tag + self._format_args(attrs) + '>')
            return

        self.xml_state.append(orig_tag)
        self.xml_line_stack.append(self.line)
        self.xml_buffer.append(unicode(''))

    def handle_endtag(self, tag):

        tag = tag.lower()
        if tag == 'xul:script':
            tag = 'script'

        if tag == 'script' and len(self.xml_buffer[-1]) > 1000:
            self.err.warning(('markup', 'complex_script'),
                             'Long inline script',
                             'Please store complex scripts in .js files '
                             'rather than inline script nodes.',
                             self.filename,
                             line=self.line,
                             context=self.context,
                             tier=2)

        if DEBUG:  # pragma: no cover
            print 'E: ', tag, self.xml_state

        if not self.xml_state:
            if 'closing_tags' in self.reported or not self.strict:
                if DEBUG:
                    print 'Unstrict; extra closing tags ------'
                return
            self.err.warning(('markup',
                              'endtag',
                              'extra_closing_tags'),
                             'Markup parsing error',
                             'The markup file has more closing tags than it '
                             'has opening tags.',
                             self.filename,
                             line=self.line,
                             context=self.context,
                             tier=2)
            self.reported.add('closing_tags')
            if DEBUG:  # pragma: no cover
                print 'Too many closing tags ------'
            return

        elif 'script' in self.xml_state[:-1]:
            # If we're in a script tag, nothing else matters. Just rush
            # everything possible into the xml buffer.

            self._save_to_buffer('</' + tag + '>')
            if DEBUG:
                print 'Markup as text in script ------'
            return

        elif tag not in self.xml_state:
            # If the tag we're processing isn't on the stack, then
            # something is wrong.
            self.err.warning(('markup',
                              'endtag',
                              'extra_closing_tags'),
                             'Parse error: tag closed before opened',
                             ['Markup tags cannot be closed before they are '
                              'opened. Perhaps you were just a little '
                              'overzealous with forward-slashes?',
                              'Tag "%s" closed before it was opened' % tag],
                             self.filename,
                             line=self.line,
                             context=self.context,
                             tier=2)
            if DEBUG:  # pragma: no cover
                print 'Tag closed before opened ------'
            return

        data_buffer = self.xml_buffer.pop()
        old_state = self.xml_state.pop()
        old_line = self.xml_line_stack.pop()
        script_type = True
        if old_state == 'script':
            script_type = self.xml_state_scripts.pop()

        # If the tag on the stack isn't what's being closed and it also
        # classifies as a self-closing tag, we just recursively close
        # down to the level of the tag we're actualy closing.
        if old_state != tag and old_state in SELF_CLOSING_TAGS:
            if DEBUG:
                print 'Self closing tag cascading down ------'
            return self.handle_endtag(tag)

        # If this is an XML-derived language, everything must nest
        # properly. No overlapping tags.
        if (old_state != tag and
            self.extension[0] == 'x' and
            not self.strict):

            self.err.warning(('markup',
                              'endtag',
                              'invalid_nesting'),
                             'Markup invalidly nested',
                             'It has been determined that the document '
                             'invalidly nests its tags. This is not permitted '
                             'in the specified document type.',
                             self.filename,
                             line=self.line,
                             context=self.context,
                             tier=2)
            if DEBUG:  # pragma: no cover
                print 'Invalid markup nesting ------'

        data_buffer = data_buffer.strip()

        # Perform analysis on collected data.
        if data_buffer:
            if tag == 'script' and not script_type:
                scripting.test_js_snippet(err=self.err, data=data_buffer,
                                          filename=self.filename,
                                          line=old_line, context=self.context)
            elif tag == 'style':
                csstester.test_css_file(self.err, self.filename, data_buffer,
                                        old_line)

    def handle_data(self, data):
        self._save_to_buffer(data)

    def handle_comment(self, data):
        self._save_to_buffer(data)

    # This code comes from markupbase.
    def parse_marked_section(self, i, report=0):
        rawdata = self.rawdata

        assert rawdata[i:i + 3] == '<![', \
               'unexpected call to parse_marked_section()'

        sectName, j = self._scan_name(i + 3, i)
        if j < 0:  # pragma: no cover
            return j
        if sectName in ('temp', 'cdata', 'ignore', 'include', 'rcdata'):
            # look for standard ]]> ending
            match = _markedsectionclose.search(rawdata, i + 3)
        else:  # pragma: no cover
            self.error('unknown status keyword %r in marked section' %
                       rawdata[i + 3:j])
        if not match:  # pragma: no cover
            return -1
        if report:  # pragma: no cover
            j = match.start(0)
            self.unknown_decl(rawdata[i + 3: j])
        return match.end(0)

    def _save_to_buffer(self, data):
        """Save data to the XML buffer for the current tag."""

        # We're not interested in data that isn't in a tag.
        if not self.xml_buffer:
            return

        data = unicodehelper.decode(data)

        self.xml_buffer[-1] += data

    def _format_args(self, args):
        """Formats a dict of HTML attributes to be in HTML attribute
        format."""

        if not args:
            return ''

        output = []
        for attr in args:
            output.append(attr[0] + '="' + attr[1] + '"')

        return ' ' + ' '.join(output)

    def _is_url_local(self, url):
        url = url.lower()
        if url.startswith('chrome://'):
            return True
        return not REMOTE_URL_PATTERN.match(url)



import json
import os
import sys
import unittest
from cStringIO import StringIO

from validator.validate import validate


FIREFOX_GUID = '{ec8030f7-c20a-464f-9b0e-13a3a9e97384}'
MOBILE_GUID = '{a23983c0-fd0e-11dc-95ff-0800200c9a66}'
THUNDERBIRD_GUID = '{3550f703-e582-4d05-9a08-453d09bdfdc6}'


def _validator(file_path, for_appversions=None, overrides=None):
    from validator.testcases import scripting
    import validator
    import validator.constants
    js = os.environ.get('SPIDERMONKEY_INSTALLATION', 'js')
    apps = os.path.join(os.path.dirname(validator.__file__),
                        'app_versions.json')
    if not os.path.exists(apps):
        raise EnvironmentError('Could not locate app_versions.json in git '
                               'repo for validator. Tried: %s' % apps)
    orig = sys.stderr
    sys.stderr = StringIO()
    try:
        result = validate(file_path, format='json',
                          # Test all tiers at once. This will make
                          # sure we see all error messages.
                          determined=True,
                          approved_applications=apps,
                          spidermonkey=js,
                          for_appversions=for_appversions,
                          timeout=60 * 3,  # seconds
                          overrides=overrides)
        sys.stdout.write(sys.stderr.getvalue())
        if 'Traceback' in sys.stderr.getvalue():
            # the validator catches and ignores certain errors in an attempt
            # to remain versatile.  There should not be any exceptions
            # while testing.
            raise RuntimeError(
                "An exception was raised during validation. Check stderr")
    finally:
        sys.stderr = orig
    return result


_cached_validation = {}


class ValidatorTest(unittest.TestCase):

    def setUp(self):
        self.validation = None
        self.messages = None
        self.ids = None

    def msg_set(self, d):
        return sorted(set([m['message'] for m in d['messages']]))

    def id_set(self, d):
        return set([tuple(m['id']) for m in d['messages']])

    def validate(self, xpi, **validate_kwargs):
        self.validation = self._run_validation(xpi, **validate_kwargs)
        self.messages = self.msg_set(self.validation)
        self.ids = self.id_set(self.validation)
        return self.validation

    def _cache_key(self, *vals):
        args = []
        for v in vals:
            self._flatten(args, v)
        return tuple(sorted(args))

    def _flatten(self, args, val):
        if isinstance(val, dict):
            for k, v in val.iteritems():
                self._flatten(args, k)
                self._flatten(args, v)
        elif isinstance(val, list):
            for v in val:
                self._flatten(args, v)
        else:
            args.append(val)

    def _run_validation(self, xpi, **validate_kwargs):
        path = os.path.join(os.path.dirname(__file__), 'addons', xpi)
        cache_key = self._cache_key(path, validate_kwargs)
        if cache_key in _cached_validation:
            return _cached_validation[cache_key]
        v = json.loads(_validator(path, **validate_kwargs))
        _cached_validation[cache_key] = v
        return v

    def assertPartialMsg(self, partial_msg):
        found = False
        for m in self.messages:
            if m.startswith(partial_msg):
                found = True
        assert found, ('Unexpected: %r' % self.messages)

    def expectMsg(self, msg):
        assert msg in self.messages, (
                    'Expected %r but only got %r' % (msg, self.messages))

    def shouldNotGetMsg(self, msg):
        assert msg not in self.messages, ('Did not expect %r' % (msg))

    def expectId(self, id):
        assert id in self.ids, (
                    'Expected %r but only got %r' % (id, self.ids))


class CompatValidatorTest(ValidatorTest):

    def validate_for_appver(self, xpi, app_guid, app_ver):
        overrides = {'targetapp_maxVersion': {app_guid: app_ver}}
        return self.validate(xpi, overrides=overrides,
                             for_appversions={app_guid: [app_ver]})


class JavaScriptTests(ValidatorTest):

    def test_createelement__used(self):
        self.validate('glee-20101227219.xpi')
        self.assertPartialMsg('createElement() used to create script tag')

    def test_dangerous_global(self):
        self.validate('feedly-addon-201101111013.xpi')
        self.expectMsg(u"`setTimeout` called in potentially "
                       u"dangerous manner")

    def test_global_called(self):
        self.validate('babuji-20110124355.xpi')
        self.expectMsg(u"`setTimeout` called in potentially "
                       u"dangerous manner")

    def test_potentially_malicious(self):
        self.validate('add-on201101101027.xpi')
        self.expectMsg(u'DOM Mutation Events Prohibited')

    def test_variable_element(self):
        self.validate('glee-20101227219.xpi')
        self.expectMsg(u'Variable element type being created')


class GeneralTests(ValidatorTest):

    def test_contains_jar_files(self):
        self.validate('test-theme-3004.jar')
        self.expectMsg(u'Add-on contains JAR files, no <em:unpack>')

    def test_potentially_illegal_name(self):
        self.validate('add-on20110110322.xpi')
        self.expectMsg(u'Add-on has potentially illegal name.')

    def test_banned_element(self):
        self.validate('gabbielsan_tools-1.01-ff.xpi')
        self.expectMsg(u'Banned element in install.rdf')

    def test_blacklisted_file(self):
        self.validate('babuji-20110124355.xpi')
        self.expectMsg(u'Flagged file extensions found.')

    def test_blacklisted_file_2(self):
        self.validate('peerscape-3.1.5-fx.xpi')
        self.expectMsg(u'Flagged file type found')

    def test_em_type_not(self):
        self.validate('babuji-20110124355.xpi')
        self.expectMsg(u'No <em:type> element found in install.rdf')

    def test_obsolete_element(self):
        self.validate('gabbielsan_tools-1.01-ff.xpi')
        self.expectMsg(u'Banned element in install.rdf')

    def test_unknown_file(self):
        self.validate('gabbielsan_tools-1.01-ff.xpi')
        self.expectMsg(u'Unrecognized element in install.rdf')

    def test_unrecognized_element(self):
        self.validate('littlemonkey-1.8.56-sm.xpi')
        self.expectMsg(u'Add-on missing manifest.')

    def test_invalid_id(self):
        self.validate('add-ongoogle-201101121132.xpi')
        self.expectMsg(u'The value of <em:id> is invalid')

    def test_xpi_cannot(self):
        self.validate('lavafox_test-theme-20101130538.xpi')
        self.expectMsg(u'Corrupt ZIP file')

    def test_invalid_version(self):
        self.validate('invalid maximum version number.xpi')
        self.expectMsg(u'Invalid maximum version number')

    def test_non_ascii_html_markup(self):
        # should be no Unicode errors
        self.validate('non-ascii-html.xpi')

    def test_webextension_seen_as_extension(self):
        validation = self.validate('beastify.xpi')
        assert validation['detected_type'] == 'extension'
        assert validation['errors'] == 0

    def test_install_rdf_and_manifest_json(self):
        validation = self.validate('installrdf-and-manifestjson.xpi')
        assert validation['detected_type'] == 'extension'
        assert validation['errors'] == 0


class LocalizationTests(ValidatorTest):

    def test_translation(self):
        self.validate('babuji-20110124355.xpi')
        self.expectMsg(u'Unchanged translation entities')

    def test_encodings(self):
        self.validate('babuji-20110124355.xpi')
        self.expectMsg(u'Unexpected encodings in locale files')

    def test_missing_translation(self):
        self.validate('download_statusbar-0.9.7.2-fx (1).xpi')
        self.expectMsg(u'Missing translation entity')


class SecurityTests(CompatValidatorTest):

    def test_missing_comments(self):
        self.validate('add-on-20110113408.xpi')
        self.expectMsg(u'Global variable overwrite')

    def test_typeless_iframes_browsers(self):
        self.validate('add-on201101081038.xpi')
        self.expectMsg(u'Typeless iframes/browsers must be local.')

    def test_binary_files(self):
        self.validate_for_appver('cooliris-1.12.2.44172-fx-mac.xpi.xpi',
                                 FIREFOX_GUID, '5.0a2')
        self.expectMsg(u"Flagged file extensions found.")
        self.expectMsg(u"Flagged file type found")
        self.expectId(('testcases_packagelayout',
                       'test_compatibility_binary',
                       'disallowed_file_type'))

    def test_thunderbird_binary_files(self):
        self.validate_for_appver('enigmail-1.2-sm-windows.xpi',
                                 THUNDERBIRD_GUID, '6.0a1')
        self.expectMsg(u"Flagged file extensions found.")
        self.expectId(('testcases_packagelayout',
                       'test_compatibility_binary',
                       'disallowed_file_type'))


class NoErrorsExpected(ValidatorTest):

    def test_an_attempt(self):
        d = self.validate('tmp.xpi')
        assert d['errors'] == 0

    def test_don_t_freak(self):
        d = self.validate('test (1).xpi')
        assert d['errors'] == 0

    def test_don_t_freak_2(self):
        d = self.validate('littlemonkey-1.8.56-sm.xpi')
        msg = self.msg_set(d)
        ok = True
        for m in msg:
            if 'install.js' in msg:
                ok = False
        assert ok, ('Unexpected: %r' % msg)

    def test_unknown_file(self):
        d = self.validate('add-on20101228444 (1).jar')
        assert d['errors'] == 0

    def test_chromemanifest_traceback(self):
        d = self.validate('chromemanifest-traceback.jar')
        assert d['errors'] == 0


class SearchTools(ValidatorTest):

    def test_opensearch_providers(self):
        self.validate('sms_search-20110115 .xml')
        self.expectMsg(u'OpenSearch: <Url> elements may not be rel=self')

    def test_opensearch_shortname(self):
        self.validate('lexisone_citation_search-20100116 .xml')
        self.expectMsg(u'OpenSearch: <ShortName> element too long')

    def test_too_many(self):
        self.validate('addon-12201-latest.xml')
        self.expectMsg(u'OpenSearch: Too many <ShortName> elements')


class JetpackDetection(ValidatorTest):

    def test_jetpack_one_signing(self):
        results = self.validate('jetpack-one.xpi')
        assert results['signing_summary']['high'] == 0

    def test_jetpack_one_jetpack(self):
        results = self.validate('jetpack-one.xpi')
        assert results['metadata']['jetpack_identified_files']

    def test_jetpack_two_signing(self):
        results = self.validate('jetpack-two.xpi')
        assert results['signing_summary']['high'] == 0

    def test_jetpack_two_jetpack(self):
        results = self.validate('jetpack-two.xpi')
        assert results['metadata']['jetpack_identified_files']

    def test_jetpack_fail_for_cfx_usage(self):
        self.validate('cfx.xpi')
        self.expectMsg(u'Add-ons built with "cfx" are no longer accepted.')



from validator import decorator
from validator.constants import PACKAGE_THEME
from validator import metadata_helpers


OPTIONS_TYPE_VALUES = ('1', '2', '3')


@decorator.register_test(tier=1)
def test_install_rdf_params(err, xpi_package=None):
    """Tests to make sure that some of the values in install.rdf
    are not gummed up."""

    if not err.get_resource('has_install_rdf'):
        return

    # We skip over install.rdf tests during bulk validation. See bug 735841.
    if err.get_resource('is_compat_test'):
        return

    install = err.get_resource('install_rdf')

    # This returns for testing reasons
    return _test_rdf(err, install)


def _test_rdf(err, install):
    """Wrapper for install.rdf testing to make unit testing so much
    easier."""

    shouldnt_exist = {'hidden'}
    if err.get_resource('listed'):
        shouldnt_exist |= {'updateURL', 'updateKey'}
    obsolete = {'file', 'skin', 'requires'}
    must_exist_once = {'id',
                       'version',
                       'name',
                       'targetApplication'}

    may_exist_once = {'about',  # For <Description> element
                      'bootstrap',
                      'optionsURL',
                      'aboutURL',
                      'iconURL',
                      'icon64URL',
                      'homepageURL',
                      'creator',
                      'multiprocessCompatible',
                      'optionsType',
                      'type',
                      'updateKey',
                      'updateURL',
                      'skinnable',
                      'strictCompatibility',
                      'unpack'}  # This has other rules; CAUTION!

    # Support a name requirement override.
    if err.overrides and err.overrides.get('ignore_empty_name'):
        must_exist_once.remove('name')
        may_exist_once.add('name')

    may_exist = {'targetApplication',
                 'localized',
                 'description',
                 'creator',
                 'translator',
                 'contributor',
                 'targetPlatform',
                 'requires',
                 'developer'}

    if (err.detected_type == PACKAGE_THEME or
        (err.subpackages and
         err.subpackages[0]['detected_type'] == PACKAGE_THEME)):
        must_exist_once.add('internalName')

    top_id = install.get_root_subject()

    predicates = {}

    for pred_raw in install.rdf.predicates(top_id, None):
        predicate = pred_raw.split('#')[-1]

        value = install.get_object(top_id, pred_raw)
        predicates[predicate] = value

        # Mark that the unpack element has been supplied
        if predicate == 'unpack':
            err.save_resource('em:unpack', value, pushable=True)

        elif predicate == 'bootstrap':
            err.save_resource('em:bootstrap', value)
            err.metadata['bootstrapped'] = value == 'true'

        elif predicate == 'multiprocessCompatible':
            err.save_resource('is_multiprocess_compatible', value == 'true')

        # Test if the predicate is banned
        elif predicate in shouldnt_exist:
            err.error(('testcases_installrdf',
                       '_test_rdf',
                       'shouldnt_exist'),
                      'Banned element in install.rdf',
                      """The element "%s" was found in the add-on's
                      install.rdf file. It is not allowed in add-ons under
                      the current configuration.""" % predicate,
                      'install.rdf')
            continue

        # Test if the predicate is obsolete
        elif predicate in obsolete:
            err.notice(('testcases_installrdf',
                        '_test_rdf',
                        'obsolete'),
                       'Obsolete element in install.rdf',
                       "The element \"%s\" was found in the add-on's "
                       'install.rdf file. It has not been banned, but it is '
                       'no longer supported by any modern Mozilla product. '
                       'Removing the element is recommended and will not '
                       'break support.' % predicate,
                       'install.rdf')
            continue

        # Remove the predicate from must_exist_once if it's there.
        if predicate in must_exist_once:
            # Test the predicate for specific values.
            if predicate in PREDICATE_TESTS:
                PREDICATE_TESTS[predicate](err, value, source='install.rdf')

            must_exist_once.remove(predicate)
            continue

        # Do the same for may_exist_once.
        elif predicate in may_exist_once:
            if (predicate == 'optionsType' and
                    str(value) not in OPTIONS_TYPE_VALUES):
                err.warning(
                    err_id=('testcases_installrdf', '_test_rdf',
                            'optionsType'),
                    warning='<em:optionsType> has bad value.',
                    description=('The value of <em:optionsType> must be one '
                                 'of: %s.' % ', '.join(OPTIONS_TYPE_VALUES),
                                 'Value found: %s' % value),
                    filename='install.rdf')

            may_exist_once.remove(predicate)
            continue

        # If the element is safe for repetition, continue
        elif predicate in may_exist:
            continue

        # If the predicate isn't in any of the above lists, it is
        # invalid and needs to go.
        err.notice(('testcases_installrdf',
                    '_test_rdf',
                    'unrecognized'),
                   'Unrecognized element in install.rdf',
                   ['An element was found in the install manifest, however it '
                    'does not appear to be a part of the specification, it '
                    'has been used too many times, or is not applicable to '
                    'the current configuration.',
                    'Detected element: <em:%s>' % predicate],
                   'install.rdf')

    if not err.get_resource('listed'):
        if 'updateURL' in predicates:
            if not (str(predicates['updateURL']).lower().startswith('https:')
                    or 'updateKey' in predicates):
                # TODO: Validate updateKey value
                err.notice(('testcases_installrdf',
                            '_test_rdf',
                            'missing_updateKey'),
                           'Missing updateKey element',
                           'Your updateURL is not served over a secure '
                           'connection, and your install.rdf does not '
                           'specify an update key. This means that serving '
                           'updates for this version will not be possible.',
                           'install.rdf',
                           signing_severity='trivial')
        else:
            err.notice(('testcases_installrdf',
                        '_test_rdf',
                        'missing_updateURL'),
                       'Missing updateURL element',
                       'Your add-on does not specify an update URL. This '
                       'means that it will be impossible for you to serve '
                       'updates to this add-on which are not listed publicly '
                       'on addons.mozilla.org.',
                       'install.rdf',
                       signing_severity='trivial')

    # Once all of the predicates have been tested, make sure there are
    # no mandatory elements that haven't been found.
    if must_exist_once:
        err.error(('testcases_installrdf',
                   '_test_rdf',
                   'missing_addon'),
                  'install.rdf missing element(s).',
                  ['The element listed is a required element in the install '
                   'manifest specification. It must be added to your addon.',
                   'Missing elements: %s' % ', '.join(must_exist_once)],
                  'install.rdf')


PREDICATE_TESTS = {
    'id': metadata_helpers.validate_id,
    'version': metadata_helpers.validate_version,
    'name': metadata_helpers.validate_name,
}



"""
Prototype
---------

args
    the raw list of arguments
traverser
    the traverser
node
    the current node being evaluated
"""

import actions
from instanceproperties import _set_HTML_property
from validator.constants import BUGZILLA_BUG


def addEventListener(args, traverser, node, wrapper):
    """
    Handle calls to addEventListener and make sure that the fourth argument is
    falsey.
    """

    if not args or len(args) < 4:
        return

    fourth_arg = traverser._traverse_node(args[3])
    if fourth_arg.get_literal_value():
        traverser.err.notice(
            err_id=('js', 'instanceactions', 'addEventListener_fourth'),
            notice='`addEventListener` called with truthy fourth argument.',
            description='When called with a truthy forth argument, listeners '
                        'can be triggered potentially unsafely by untrusted '
                        'code. This requires careful review.',
            filename=traverser.filename,
            line=traverser.line,
            column=traverser.position,
            context=traverser.context)


def createElement(args, traverser, node, wrapper):
    """Handles createElement calls"""

    if not args:
        return

    simple_args = map(traverser._traverse_node, args)

    first_as_str = actions._get_as_str(simple_args[0].get_literal_value())
    if first_as_str.lower() == u'script':
        _create_script_tag(traverser)
    elif not simple_args[0].is_literal():
        _create_variable_element(traverser)


def createElementNS(args, traverser, node, wrapper):
    """Handles createElementNS calls"""

    if not args or len(args) < 2:
        return

    simple_args = map(traverser._traverse_node, args)

    second_as_str = actions._get_as_str(simple_args[1].get_literal_value())
    if 'script' in second_as_str.lower():
        _create_script_tag(traverser)
    elif not simple_args[1].is_literal():
        _create_variable_element(traverser)


def QueryInterface(args, traverser, node, wrapper):
    """Handles QueryInterface calls"""

    if not args:
        return

    from call_definitions import xpcom_constructor
    return xpcom_constructor('QueryInterface', True, True)(
        wrapper=node,
        arguments=args,
        traverser=traverser)


def getInterface(args, traverser, node, wrapper):
    """Handles getInterface calls"""

    # This really only needs to be handled for nsIInterfaceRequestor
    # intarfaces, but as it's fair for code to assume that that
    # interface has already been queried and methods with this name
    # are unlikely to behave differently, we just process it for all
    # objects.

    if not args:
        return

    from call_definitions import xpcom_constructor
    return xpcom_constructor('getInterface')(
        wrapper=node,
        arguments=args,
        traverser=traverser)


def _create_script_tag(traverser):
    """Raises a warning that the dev is creating a script tag"""
    traverser.warning(
        err_id=('testcases_javascript_instanceactions', '_call_expression',
                'called_createelement'),
        warning='createElement() used to create script tag',
        description='Dynamic creation of script nodes can be unsafe if '
                    'contents are not static or are otherwise unsafe, '
                    'or if `src` is remote.',
        signing_help='Please avoid using <script> tags to load scripts. '
                     'For potential alternatives, please see '
                     'https://developer.mozilla.org/en-US/Add-ons/'
                     'Overlay_Extensions/XUL_School/'
                     'Appendix_D:_Loading_Scripts',
        signing_severity='medium')


def _create_variable_element(traverser):
    """Raises a warning that the dev is creating an arbitrary element"""
    traverser.err.warning(
        err_id=('testcases_javascript_instanceactions', '_call_expression',
                'createelement_variable'),
        warning='Variable element type being created',
        description=('createElement or createElementNS were used with a '
                     'variable rather than a raw string. Literal values '
                     'should be used when taking advantage of the element '
                     'creation functions.',
                     "E.g.: createElement('foo') rather than "
                     'createElement(el_type)'),
        filename=traverser.filename,
        line=traverser.line,
        column=traverser.position,
        context=traverser.context)


def setAttribute(args, traverser, node, wrapper):
    """This ensures that setAttribute calls don't set on* attributes"""

    if not args:
        return

    simple_args = [traverser._traverse_node(a) for a in args]

    first_as_str = actions._get_as_str(simple_args[0].get_literal_value())
    if first_as_str.lower().startswith('on'):
        traverser.warning(
            err_id=('testcases_javascript_instanceactions', 'setAttribute',
                    'setting_on*'),
            warning='on* attribute being set using setAttribute',
            description='To prevent vulnerabilities, event handlers (like '
                        "'onclick' and 'onhover') should always be defined "
                        'using addEventListener.',
            signing_help='Please use `addEventListener` any place you might '
                         'otherwise create event listener attributes. Event '
                         'listener attributes will not be accepted in add-ons '
                         'submitted for automated signing in any instance '
                         'where they may be reasonably avoided.',
            signing_severity='medium')


def insertAdjacentHTML(args, traverser, node, wrapper):
    """
    Perfrom the same tests on content inserted into the DOM via
    insertAdjacentHTML as we otherwise would for content inserted via the
    various innerHTML/outerHTML properties.
    """
    if not args or len(args) < 2:
        return

    content = traverser._traverse_node(args[1])
    _set_HTML_property('insertAdjacentHTML', content, traverser)


def launch(args, traverser, node, wrapper):
    traverser.warning(
        err_id=('testcases_javascript_instanceactions', 'launch'),
        warning='Potentially dangerous use of `launch()`',
        description='Use of the `nsIFile.launch()` method can be dangerous, '
                    'and requires careful review.',
        editors_only=True)


def openDialog(args, traverser, node, wrapper):
    """Raise an error if the first argument is a remote URL."""
    if not args:
        return
    uri = traverser._traverse_node(args[0])
    from call_definitions import open_in_chrome_context
    open_in_chrome_context(uri, 'openDialog', traverser)


def bind(args, traverser, node, wrapper):
    if 'callee' not in node and 'object' not in node['callee']:
        return
    obj = traverser._traverse_node(node['callee']['object'])
    if obj.callable:
        return obj


SYNCHRONOUS_SQL_DESCRIPTION = (
    'The use of synchronous SQL via the storage system leads to severe '
    'responsiveness issues, and should be avoided at all costs. Please '
    'use asynchronous SQL via Sqlite.jsm (http://mzl.la/sqlite-jsm) or '
    'the `executeAsync` method, or otherwise switch to a simpler database '
    'such as JSON files or IndexedDB.')


def _check_dynamic_sql(args, traverser, node=None, wrapper=None):
    """
    Checks for the use of non-static strings when creating/exeucting SQL
    statements.
    """

    simple_args = map(traverser._traverse_node, args)
    if len(args) >= 1 and not simple_args[0].is_literal():
        traverser.warning(
            err_id=('js', 'instanceactions', 'executeSimpleSQL_dynamic'),
            warning='SQL statements should be static strings',
            description=('Dynamic SQL statement should be constucted via '
                         'static strings, in combination with dynamic '
                         'parameter binding via Sqlite.jsm wrappers '
                         '(http://mzl.la/sqlite-jsm) or '
                         '`createAsyncStatement` '
                         '(https://developer.mozilla.org/en-US/docs'
                         '/Storage#Binding_parameters)'),
            filename=traverser.filename,
            line=traverser.line,
            column=traverser.position,
            context=traverser.context)


def createStatement(args, traverser, node, wrapper):
    """
    Handles calls to `createStatement`, returning an object which emits
    warnings upon calls to `execute` and `executeStep` rather than
    `executeAsync`.
    """
    _check_dynamic_sql(args, traverser)
    from predefinedentities import build_quick_xpcom
    return build_quick_xpcom('createInstance', 'mozIStorageBaseStatement',
                             traverser, wrapper=True)


def executeSimpleSQL(args, traverser, node, wrapper):
    """
    Handles calls to `executeSimpleSQL`, warning that asynchronous methods
    should be used instead.
    """
    _check_dynamic_sql(args, traverser)
    traverser.err.warning(
        err_id=('js', 'instanceactions', 'executeSimpleSQL'),
        warning='Synchronous SQL should not be used',
        description=SYNCHRONOUS_SQL_DESCRIPTION,
        filename=traverser.filename,
        line=traverser.line,
        column=traverser.position,
        context=traverser.context)


def create_preference_branch(arguments, traverser, node, wrapper):
    """Creates a preference branch, which can be used for testing composed
    preference names."""

    if arguments:
        arg = traverser._traverse_node(arguments[0])
        if arg.is_literal():
            # Avoid import loop
            from predefinedentities import build_quick_xpcom
            res = build_quick_xpcom('createInstance', 'nsIPrefBranch',
                                    traverser, wrapper=True)
            res.value['preference_branch'] = actions._get_as_str(arg)
            return res


def drop_pref_messages(wrapper):
    """Drop any preference-related messages for the given wrapper, if that
    wrapper is an immediate literal that was passed as an argument, and the
    messages are on the same line as the traverser.

    Used to ignore preference warnings when the strings are provably being
    read rather than written, or when they're provably being written and
    have a more useful, redundant warning already.
    """

    traverser = wrapper.traverser

    if wrapper.value.source == 'arguments':
        # Avoid import loop.
        from validator.testcases.regex import PREFERENCE_ERROR_ID

        for msg in wrapper.value.messages:
            if (msg['id'] == PREFERENCE_ERROR_ID and
                    (msg['file'], msg['line']) == (
                        traverser.filename, traverser.line)):
                traverser.err.drop_message(msg)


def get_preference(wrapper, arguments, traverser):
    """Tests get preference calls, and removes preference write warnings
    when they are not necessary."""

    if len(arguments) >= 1:
        arg = traverser._traverse_node(arguments[0])
        if arg.is_literal():
            drop_pref_messages(arg)


def set_preference(wrapper, arguments, traverser):
    """Tests set preference calls for non-root preferences branches against
    dangerous values."""

    if len(arguments) < 1:
        return

    parent = getattr(wrapper, 'parent', None)
    arg = traverser._traverse_node(arguments[0])
    if arg.is_literal():
        pref = actions._get_as_str(arg)

        # If we're being called on a preference branch other than the root,
        # prepend its branch name to the passed preference name.
        if (parent and parent.is_global and
                parent.value.get('preference_branch')):
            pref = parent.value['preference_branch'] + pref
        else:
            drop_pref_messages(arg)

        kw = {'err_id': ('testcases_javascript_actions',
                         '_call_expression', 'called_set_preference'),
              'warning': 'Attempt to set a dangerous preference'}

        # Local import to prevent import loop.
        from validator.testcases.regex import validate_pref
        validate_pref(pref, traverser, kw, wrapper=arg)


INSTANCE_DEFINITIONS = {
    'addEventListener': addEventListener,
    'bind': bind,
    'createElement': createElement,
    'createElementNS': createElementNS,
    'createAsyncStatement': _check_dynamic_sql,
    'createStatement': createStatement,
    'executeSimpleSQL': executeSimpleSQL,
    'getBranch': create_preference_branch,
    'getDefaultBranch': create_preference_branch,
    'getInterface': getInterface,
    'insertAdjacentHTML': insertAdjacentHTML,
    'launch': launch,
    'openDialog': openDialog,
    'QueryInterface': QueryInterface,
    'setAttribute': setAttribute,
}



from helper import CompatTestCase
from validator.compat import FX47_DEFINITION


class TestFX47Compat(CompatTestCase):
    """Test that compatibility tests for Gecko 47 are properly executed."""

    VERSION = FX47_DEFINITION

    def test_bnsIX509CertDB_method_changed(self):
        """https://github.com/mozilla/addons-server/issues/2221"""
        expected = (
            'Most methods in nsIX509CertDB had their unused '
            'arguments removed.')

        script = '''
            var certDB = Cc["@mozilla.org/security/x509certdb;1"]
                    .getService(Ci.nsIX509CertDB);
            certDB.importPKCS12File(null, nsIFile);
        '''

        self.run_script_for_compat(script)

        assert not self.compat_err.notices
        assert not self.compat_err.errors

        # There are other checks because using nsIX509CertDB is
        # potentially dangerous.
        warning = [x for x in self.compat_err.warnings if x['message'] == expected]
        assert warning
        assert warning[0]['compatibility_type'] == 'error'

        script = '''
            var certDB = Cc["@mozilla.org/security/x509certdb;1"]
                                        .getService(Ci.nsIX509CertDB);
            certdb.importCertsFromFile(null, fp.file, nsIX509Cert.CA_CERT);
        '''

        self.run_script_for_compat(script)

        assert not self.compat_err.notices
        assert not self.compat_err.errors

        # There are other checks because using nsIX509CertDB is
        # potentially dangerous.
        warning = [x for x in self.compat_err.warnings if x['message'] == expected]
        assert warning
        assert warning[0]['compatibility_type'] == 'error'


        script = 'const nsIX509CertDB = Components.interfaces.nsIX509CertDB;'

        self.run_script_for_compat(script)

        assert not self.compat_err.notices
        assert not self.compat_err.errors

        # There are other checks because using nsIX509CertDB is
        # potentially dangerous.
        warning = [x for x in self.compat_err.warnings if x['message'] == expected]
        assert warning
        assert warning[0]['compatibility_type'] == 'error'

    def test_changed_return_type(self):
        """https://github.com/mozilla/addons-server/issues/2220"""
        message = (
            'listTokens(), listModules() and listSlots() now return '
            'nsISimpleEnumerator instead of nsIEnumerator.')

        script = '''
            var db = Components.classes["@mozilla.org/security/pk11tokendb;1"]
                        .getService(Components.interfaces.nsIPK11TokenDB);
            var modules = db.listTokens();
        '''
        self.run_script_for_compat(script)

        assert not self.compat_err.notices
        assert not self.compat_err.errors

        assert len(self.compat_err.warnings) == 1
        assert self.compat_err.warnings[0]['compatibility_type'] == 'error'
        assert self.compat_err.warnings[0]['message'].startswith(message)

        script = '''
            var db = Components.classes["@mozilla.org/security/pkcs11moduledb;1"]
                        .getService(Components.interfaces.nsIPKCS11ModuleDB);
            var modules = db.listModules();
        '''
        self.run_script_for_compat(script)

        assert not self.compat_err.notices
        assert not self.compat_err.errors

        assert len(self.compat_err.warnings) == 1
        assert self.compat_err.warnings[0]['compatibility_type'] == 'error'
        assert self.compat_err.warnings[0]['message'].startswith(message)

        script = '''
            var modules = modules.QueryInterface(Components.interfaces.nsIPKCS11Module);
            modules.listSlots();
        '''
        self.run_script_for_compat(script)

        assert not self.compat_err.notices
        assert not self.compat_err.errors

        assert len(self.compat_err.warnings) == 1
        assert self.compat_err.warnings[0]['compatibility_type'] == 'error'
        assert self.compat_err.warnings[0]['message'].startswith(message)

        script = 'var modules = mylib.listModules();'
        self.run_script_for_compat(script)

        assert not self.compat_err.notices
        assert not self.compat_err.errors
        assert not self.compat_err.warnings

    def test_gDevTools_moved(self):
        """https://github.com/mozilla/addons-server/issues/2219"""
        message = 'The gDevTools.jsm module should no longer be used.'

        script = (
            'Components.utils.import("resource:///modules/devtools/client'
            '/framework/gDevTools.jsm", ainspectorSidebar);')

        self.run_script_for_compat(script)

        assert not self.compat_err.notices
        assert not self.compat_err.errors

        assert len(self.compat_err.warnings) == 1
        assert self.compat_err.warnings[0]['compatibility_type'] == 'warning'
        assert self.compat_err.warnings[0]['message'].startswith(message)

        # Test that we actually match on "." and not "any character"
        script = (
            'Components.utils.import("resource:///modules/devtools/client'
            '/framework/gDevTools_jsm", ainspectorSidebar);')
        self.run_script_for_compat(script)

        assert not self.compat_err.notices
        assert not self.compat_err.errors
        assert not self.compat_err.warnings

    def test_FUEL_library_removed(self):
        """https://github.com/mozilla/addons-server/issues/2218"""
        message = 'The FUEL library is no longer supported.'

        script = ('var app = Components'
                  '.classes["@mozilla.org/fuel/application;1"]'
                  '.getService(Components.interfaces.fuelIApplication);')

        self.run_script_for_compat(script)

        assert not self.compat_err.notices
        assert not self.compat_err.errors

        assert len(self.compat_err.warnings) == 1
        assert self.compat_err.warnings[0]['compatibility_type'] == 'error'
        assert self.compat_err.warnings[0]['message'].startswith(message)

        script = (
            'var nowDebug = !!Application.prefs.get'
            '("extensions.hatenabookmark.debug.log").value;')

        self.run_script_for_compat(script)

        assert not self.compat_err.notices
        assert not self.compat_err.errors

        assert len(self.compat_err.warnings) == 1
        assert self.compat_err.warnings[0]['compatibility_type'] == 'error'
        assert self.compat_err.warnings[0]['message'].startswith(message)

        # Make sure the regex isn't too wide
        script = 'var application = "something not so serious";'

        self.run_script_for_compat(script)

        assert not self.compat_err.notices
        assert not self.compat_err.errors
        assert not self.compat_err.warnings

    def test_about_customizing_removed(self):
        """https://github.com/mozilla/addons-server/issues/2217"""
        message = 'The customization panel is no longer loaded via about:customizing'

        script = (
            '// * modified from: resource:///modules/'
            'CustomizationTabPreloader.jsm')

        self.run_script_for_compat(script)

        assert not self.compat_err.notices
        assert not self.compat_err.errors

        assert len(self.compat_err.warnings) == 1
        assert self.compat_err.warnings[0]['compatibility_type'] == 'error'
        assert self.compat_err.warnings[0]['message'].startswith(message)

        script = (
            '// * ensurePreloading starts the preloading of the '
            'about:customizing')

        self.run_script_for_compat(script)

        assert not self.compat_err.notices
        assert not self.compat_err.errors

        assert len(self.compat_err.warnings) == 1
        assert self.compat_err.warnings[0]['compatibility_type'] == 'error'
        assert self.compat_err.warnings[0]['message'].startswith(message)

        script = (
            'Components.utils.import("resource:///modules/'
            'CustomizationTabPreloader.jsm");')

        self.run_script_for_compat(script)

        assert not self.compat_err.notices
        assert not self.compat_err.errors

        assert len(self.compat_err.warnings) == 1
        assert self.compat_err.warnings[0]['compatibility_type'] == 'error'
        assert self.compat_err.warnings[0]['message'].startswith(message)



from validator import decorator
from validator.constants import PACKAGE_THEME


@decorator.register_test(tier=2, expected_type=PACKAGE_THEME)
def test_theme_manifest(err, xpi_package=None):
    """Tests the chrome.manifest files in the package for
    compliance with the standard theme triples."""

    # Don't even both with the test(s) if there's no chrome.manifest.
    chrome = err.get_resource('chrome.manifest')
    if not chrome:
        return

    for triple in chrome.triples:
        subject = triple['subject']
        # Test to make sure that the triple's subject is valid
        if subject not in ('skin', 'style'):
            err.warning(
                err_id=('themes', 'test_theme_manifest',
                        'invalid_chrome_manifest_subject'),
                warning='Invalid chrome.manifest subject',
                description=('chrome.manifest files for full themes are only '
                             "allowed to have 'skin' and 'style' items. "
                             'Other types of items are disallowed for '
                             'security reasons.',
                             'Invalid subject: %s' % subject),
                filename=triple['filename'],
                line=triple['line'],
                context=triple['context'])



'Constants that will be used across files.'

import json
import os
import re
import types


DESCRIPTION_TYPES = types.StringTypes + (list, tuple, )

# Package type constants.
PACKAGE_ANY = 0
PACKAGE_EXTENSION = 1
PACKAGE_THEME = 2
PACKAGE_DICTIONARY = 3
PACKAGE_LANGPACK = 4
PACKAGE_SEARCHPROV = 5
PACKAGE_MULTI = 1  # A multi extension is an extension
PACKAGE_SUBPACKAGE = 7

# The "earliest" version number for Firefox 4
FF4_MIN = '3.7a1pre'

# Application GUIDs
FIREFOX_GUID = '{ec8030f7-c20a-464f-9b0e-13a3a9e97384}'
MOZILLA_GUID = '{86c18b42-e466-45a9-ae7a-9b95ba6f5640}'
THUNDERBIRD_GUID = '{3550f703-e582-4d05-9a08-453d09bdfdc6}'
SUNBIRD_GUID = '{718e30fb-e89b-41dd-9da7-e25a45638b28}'
SEAMONKEY_GUID = '{92650c4d-4b8e-4d2a-b7eb-24ecf4f6b63a}'
FENNEC_GUID = '{a23983c0-fd0e-11dc-95ff-0800200c9a66}'
ANDROID_GUID = '{aa3c5121-dab2-40e2-81ca-7ea25febc110}'

APPLICATIONS = {
    FIREFOX_GUID: 'firefox',
    MOZILLA_GUID: 'mozilla',
    THUNDERBIRD_GUID: 'thunderbird',
    SUNBIRD_GUID: 'sunbird',
    SEAMONKEY_GUID: 'seamonkey',
    FENNEC_GUID: 'fennec',
    ANDROID_GUID: 'android',
}

with open(os.path.join(os.path.dirname(__file__), 'app_versions.json')) as avs:
    APPROVED_APPLICATIONS = json.load(avs)

BUGZILLA_BUG = 'https://bugzil.la/%d'
MDN_DOC = 'https://developer.mozilla.org/docs/%s'

JETPACK_URI_URL = 'https://wiki.mozilla.org/Labs/Jetpack/Release_Notes/' \
                      '1.4#Known_Issues'

# The maximum size of any string in JS analysis.
MAX_STR_SIZE = 1024 * 24  # 24KB

# The maximum number of JS files that can be exhaustively validated in one
# package.
MAX_JS_THRESHOLD = 900


# Severities of signing-related messages, from least severe to most.
SIGNING_SEVERITIES = ('trivial', 'low', 'medium', 'high')

# The pattern that matches event assignments
# TODO(valcom): Move this to valcom when that's a thing.
EVENT_ASSIGNMENT = re.compile('<.+ on[a-z]+=')

try:
    from validator.constants_local import *  # noqa
except ImportError:
    pass



import json
import logging
import types
import uuid
from StringIO import StringIO

import validator
from validator import unicodehelper
from validator.constants import SIGNING_SEVERITIES
from validator.outputhandlers.shellcolors import OutputHandler


log = logging.getLogger('amo.validator')


class ErrorBundle(object):
    """This class does all sorts of cool things. It gets passed around
    from test to test and collects up all the errors like the candy man
    'separating the sorrow and collecting up all the cream.' It's
    borderline magical.

    Keyword Arguments

    **determined**
        Whether the validator should continue after a tier fails
    **listed**
        True if the add-on is destined for AMO, false if not
    **instant**
        Who knows what this does
    **overrides**
        dict of install.rdf values to override. Possible keys:
        targetapp_minVersion, targetapp_maxVersion
    **for_appversions**
        A dict of app GUIDs referencing lists of versions. Determines which
        version-dependant tests should be run.
    """

    def __init__(self, determined=True, listed=True, instant=False,
                 overrides=None, for_appversions=None):

        self.handler = None

        self.errors = []
        self.warnings = []
        self.notices = []
        self.message_tree = {}

        self.compat_summary = {'errors': 0,
                               'warnings': 0,
                               'notices': 0}
        self.signing_summary = {s: 0 for s in SIGNING_SEVERITIES}

        self.ending_tier = 1
        self.tier = 1

        self.subpackages = []
        self.package_stack = []

        self.detected_type = 0
        self.unfinished = False

        # TODO: Break off into resource helper
        self.resources = {}
        self.pushable_resources = {}
        self.final_context = None

        self.metadata = {'requires_chrome': False, 'listed': listed,
                         'validator_version': validator.__version__}
        if listed:
            self.resources['listed'] = True
        self.instant = instant
        self.determined = determined

        self.version_requirements = None

        self.overrides = overrides or None

        self.supported_versions = self.for_appversions = for_appversions

    def _message(type_, message_type):
        def wrap(self, *args, **kwargs):
            message = {
                'id': kwargs.get('err_id') or args[0],
                'message': kwargs.get(message_type) or args[1],
                'description': kwargs.get('description',
                                          args[2] if len(args) > 2 else None),
                # Filename is never None.
                'file': kwargs.get('filename',
                                   args[3] if len(args) > 3 else ''),
                'line': kwargs.get('line',
                                   args[4] if len(args) > 4 else None),
                'column': kwargs.get('column',
                                     args[5] if len(args) > 5 else None),
                # If true, the message should only be shown to editors.
                'editors_only': kwargs.get('editors_only', False),
            }
            for field in ('tier', 'for_appversions', 'compatibility_type', ):
                message[field] = kwargs.get(field)

            if 'signing_severity' in kwargs:
                severity = kwargs['signing_severity']

                assert severity in SIGNING_SEVERITIES

                if not kwargs.get('from_merge'):
                    self.signing_summary[severity] += 1
                message['signing_severity'] = severity
            if 'signing_help' in kwargs:
                message['signing_help'] = kwargs['signing_help']

            self._save_message(getattr(self, type_), type_, message,
                               from_merge=kwargs.get('from_merge'),
                               context=kwargs.get('context'))
            return message

        wrap.__name__ = message_type
        return wrap

    # And then all the real functions. Ahh, how clean!
    error = _message('errors', 'error')
    warning = _message('warnings', 'warning')
    notice = _message('notices', 'notice')

    def system_error(self, msg_id=None, message=None, description=None,
                     validation_timeout=False, exc_info=None, **kw):
        """Add an error message for an unexpected exception in validator
        code, and move it to the front of the error message list. If
        `exc_info` is supplied, the error will be logged.

        If the error is a validation timeout, it is re-raised unless
        `msg_id` is "validation_timeout"."""

        if exc_info:
            if (isinstance(exc_info[1], validator.ValidationTimeout) and
                    msg_id != 'validation_timeout'):
                # These should always propagate to the top-level exception
                # handler, and be reported only once.
                raise exc_info[1]

            log.error('Unexpected error during validation: %s: %s'
                      % (exc_info[0].__name__, exc_info[1]),
                      exc_info=exc_info)

        full_id = ('validator', 'unexpected_exception')
        if msg_id:
            full_id += (msg_id,)

        self.error(full_id,
                   message or 'An unexpected error has occurred.',
                   description or
                   ('Validation was unable to complete successfully due '
                    'to an unexpected error.',
                    'The error has been logged, but please consider '
                    'filing an issue report here: '
                    'https://bit.ly/1POrYYU'),
                   tier=1, **kw)

        # Move the error message to the beginning of the list.
        self.errors.insert(0, self.errors.pop())

    def drop_message(self, message):
        """Drop the given message object from the appropriate message list.

        Returns True if the message was found, otherwise False."""

        for type_ in 'errors', 'warnings', 'notices':
            list_ = getattr(self, type_)
            if message in list_:
                list_.remove(message)
                if 'signing_severity' in message:
                    self.signing_summary[message['signing_severity']] -= 1
                return True

        return False

    def set_tier(self, tier):
        'Updates the tier and ending tier'
        self.tier = tier
        if tier > self.ending_tier:
            self.ending_tier = tier

    @property
    def message_count(self):
        return len(self.errors) + len(self.warnings) + len(self.notices)

    def _save_message(self, stack, type_, message, context=None,
                      from_merge=False):
        'Stores a message in the appropriate message stack.'

        uid = uuid.uuid4().hex
        message['uid'] = uid

        # Get the context for the message (if there's a context available)
        if context is not None:
            if isinstance(context, tuple):
                message['context'] = context
            else:
                message['context'] = (
                    context.get_context(line=message['line'],
                                        column=message['column']))
        else:
            message['context'] = None

        if self.package_stack:
            if not isinstance(message['file'], list):
                message['file'] = [message['file']]

            message['file'] = self.package_stack + message['file']

        # Test that if for_appversions is set that we're only applying to
        # supported add-ons. THIS IS THE LAST FILTER BEFORE THE MESSAGE IS
        # ADDED TO THE STACK!
        if message['for_appversions']:
            if not self.supports_version(message['for_appversions']):
                if self.instant:
                    print '(Instant error discarded)'
                    self._print_message(type_, message, verbose=True)
                return
        elif self.version_requirements:
            # If there was no for_appversions but there were version
            # requirements detailed in the decorator, use the ones from the
            # decorator.
            message['for_appversions'] = self.version_requirements

        # Save the message to the stack.
        stack.append(message)

        # Mark the tier that the error occurred at.
        if message['tier'] is None:
            message['tier'] = self.tier

        # Build out the compatibility summary if possible.
        if message['compatibility_type'] and not from_merge:
            self.compat_summary['%ss' % message['compatibility_type']] += 1

        # Build out the message tree entry.
        if message['id']:
            tree = self.message_tree
            last_id = None
            for eid in message['id']:
                if last_id is not None:
                    tree = tree[last_id]
                if eid not in tree:
                    tree[eid] = {'__errors': 0,
                                 '__warnings': 0,
                                 '__notices': 0,
                                 '__messages': []}
                tree[eid]['__%s' % type_] += 1
                last_id = eid

            tree[last_id]['__messages'].append(uid)

        # If instant mode is turned on, output the message immediately.
        if self.instant:
            self._print_message(type_, message, verbose=True)

    def failed(self, fail_on_warnings=True):
        """Returns a boolean value describing whether the validation
        succeeded or not."""

        return bool(self.errors) or (fail_on_warnings and bool(self.warnings))

    def get_resource(self, name):
        'Retrieves an object that has been stored by another test.'

        if name in self.resources:
            return self.resources[name]
        elif name in self.pushable_resources:
            return self.pushable_resources[name]
        else:
            return False

    def save_resource(self, name, resource, pushable=False):
        'Saves an object such that it can be used by other tests.'

        if pushable:
            self.pushable_resources[name] = resource
        else:
            self.resources[name] = resource

    @property
    def is_nested_package(self):
        'Returns whether the current package is within a PACKAGE_MULTI'
        return bool(self.package_stack)

    def push_state(self, new_file=''):
        'Saves the current error state to parse subpackages'

        self.subpackages.append({'detected_type': self.detected_type,
                                 'message_tree': self.message_tree,
                                 'resources': self.pushable_resources,
                                 'metadata': self.metadata})

        self.message_tree = {}
        self.pushable_resources = {}
        self.metadata = {'requires_chrome': False,
                         'listed': self.metadata.get('listed'),
                         'validator_version': validator.__version__}

        self.package_stack.append(new_file)

    def pop_state(self):
        'Retrieves the last saved state and restores it.'

        # Save a copy of the current state.
        state = self.subpackages.pop()
        metadata = self.metadata
        # We only rebuild message_tree anyway. No need to restore.

        # Copy the existing state back into place
        self.detected_type = state['detected_type']
        self.message_tree = state['message_tree']
        self.pushable_resources = state['resources']
        self.metadata = state['metadata']

        name = self.package_stack.pop()

        self.metadata.setdefault('sub_packages', {})[name] = metadata

    def render_json(self):
        'Returns a JSON summary of the validation operation.'

        types = {0: 'unknown',
                 1: 'extension',
                 2: 'theme',
                 3: 'dictionary',
                 4: 'langpack',
                 5: 'search',
                 8: 'webapp'}
        output = {'detected_type': types[self.detected_type],
                  'ending_tier': self.ending_tier,
                  'success': not self.failed(),
                  'messages': [],
                  'errors': len(self.errors),
                  'warnings': len(self.warnings),
                  'notices': len(self.notices),
                  'message_tree': self.message_tree,
                  'compatibility_summary': self.compat_summary,
                  'signing_summary': self.signing_summary,
                  'metadata': self.metadata}

        messages = output['messages']

        # Copy messages to the JSON output
        for error in self.errors:
            error['type'] = 'error'
            messages.append(error)

        for warning in self.warnings:
            warning['type'] = 'warning'
            messages.append(warning)

        for notice in self.notices:
            notice['type'] = 'notice'
            messages.append(notice)

        # Output the JSON.
        return json.dumps(output)

    def print_summary(self, verbose=False, no_color=False):
        'Prints a summary of the validation process so far.'

        types = {0: 'Unknown',
                 1: 'Extension/Multi-Extension',
                 2: 'Full Theme',
                 3: 'Dictionary',
                 4: 'Language Pack',
                 5: 'Search Provider',
                 7: 'Subpackage',
                 8: 'App'}
        detected_type = types[self.detected_type]

        buffer = StringIO()
        self.handler = OutputHandler(buffer, no_color)

        # Make a neat little printout.
        self.handler.write('\n<<GREEN>>Summary:') \
            .write('-' * 30) \
            .write('Detected type: <<BLUE>>%s' % detected_type) \
            .write('-' * 30)

        if self.failed():
            self.handler.write('<<BLUE>>Test failed! Errors:')

            # Print out all the errors/warnings:
            for error in self.errors:
                self._print_message('<<RED>>Error:<<NORMAL>>\t',
                                    error, verbose)
            for warning in self.warnings:
                self._print_message('<<YELLOW>>Warning:<<NORMAL>> ',
                                    warning, verbose)
        else:
            self.handler.write('<<GREEN>>All tests succeeded!')

        if self.notices:
            for notice in self.notices:
                self._print_message(prefix='<<WHITE>>Notice:<<NORMAL>>\t',
                                    message=notice,
                                    verbose=verbose)

        if 'is_jetpack' in self.metadata and verbose:
            self.handler.write('\n')
            self.handler.write('<<GREEN>>Jetpack add-on detected.<<NORMAL>>\n'
                               'Identified files:')
            if 'jetpack_identified_files' in self.metadata:
                for filename, data in \
                        self.metadata['jetpack_identified_files'].items():
                    self.handler.write((' %s\n' % filename) +
                                       ('  %s : %s' % data))

            if 'jetpack_unknown_files' in self.metadata:
                self.handler.write('Unknown files:')
                for filename in self.metadata['jetpack_unknown_files']:
                    self.handler.write(' %s' % filename)

        self.handler.write('\n')
        if self.unfinished:
            self.handler.write('<<RED>>Validation terminated early')
            self.handler.write('Errors during validation are preventing '
                               'the validation process from completing.')
            self.handler.write('Use the <<YELLOW>>--determined<<NORMAL>> '
                               'flag to ignore these errors.')
            self.handler.write('\n')

        return buffer.getvalue()

    def _flatten_list(self, data):
        'Flattens nested lists into strings.'

        if data is None:
            return ''
        if isinstance(data, types.StringTypes):
            return data
        elif isinstance(data, (list, tuple)):
            return '\n'.join(self._flatten_list(x) for x in data)

    def _print_message(self, prefix, message, verbose=True):
        'Prints a message and takes care of all sorts of nasty code'

        # Load up the standard output.
        output = ['\n', prefix, message['message']]

        # We have some extra stuff for verbose mode.
        if verbose:
            verbose_output = []

            # Detailed problem description.
            if message['description']:
                verbose_output.append(
                    self._flatten_list(message['description']))

            if message.get('signing_severity'):
                verbose_output.append(
                    ('\tAutomated signing severity: %s' %
                     message['signing_severity']))

            if message.get('signing_help'):
                verbose_output.append(
                    '\tSuggestions for passing automated signing:')
                verbose_output.append(
                    self._flatten_list(message['signing_help']))

            # Show the user what tier we're on
            verbose_output.append('\tTier:\t%d' % message['tier'])

            # If file information is available, output that as well.
            files = message['file']
            if files is not None and files != '':
                fmsg = '\tFile:\t%s'

                # Nested files (subpackes) are stored in a list.
                if type(files) is list:
                    if files[-1] == '':
                        files[-1] = '(none)'
                    verbose_output.append(fmsg % ' > '.join(files))
                else:
                    verbose_output.append(fmsg % files)

            # If there is a line number, that gets put on the end.
            if message['line']:
                verbose_output.append('\tLine:\t%s' % message['line'])
            if message['column'] and message['column'] != 0:
                verbose_output.append('\tColumn:\t%d' % message['column'])

            if message.get('context'):
                verbose_output.append('\tContext:')
                verbose_output.extend([('\t> %s' % x
                                        if x is not None
                                        else '\t>' + ('-' * 20))
                                       for x
                                       in message['context']])

            # Stick it in with the standard items.
            output.append('\n')
            output.append('\n'.join(verbose_output))

        # Send the final output to the handler to be rendered.
        self.handler.write(u''.join(map(unicodehelper.decode, output)))

    def supports_version(self, guid_set):
        """
        Returns whether a GUID set in for_appversions format is compatbile with
        the current supported applications list.
        """

        # Don't let the test run if we haven't parsed install.rdf yet.
        if self.supported_versions is None:
            raise Exception('Early compatibility test run before install.rdf '
                            'was parsed.')

        return self._compare_version(requirements=guid_set,
                                     support=self.supported_versions)

    def _compare_version(self, requirements, support):
        """
        Return whether there is an intersection between a support applications
        GUID set and a set of supported applications.
        """

        for guid in requirements:
            # If we support any of the GUIDs in the guid_set, test if any of
            # the provided versions for the GUID are supported.
            if (guid in support and
                any((detected_version in requirements[guid]) for
                    detected_version in support[guid])):
                return True

    def discard_unused_messages(self, ending_tier):
        """
        Delete messages from errors, warnings, and notices whose tier is
        greater than the ending tier.
        """

        stacks = [self.errors, self.warnings, self.notices]
        for stack in stacks:
            for message in stack:
                if message['tier'] > ending_tier:
                    stack.remove(message)
